import typing
from commonfate_provider.runtime import AWSLambdaRuntime
from commonfate_provider import provider, args


class BasicProvider(provider.Provider):
    pass


def fetch_groups(provider: BasicProvider) -> typing.List[args.Option]:
    return [
        args.Option(value="one", label="one"),
        args.Option(value="two", label="two"),
        args.Option(value="three", label="three"),
    ]


class Args(args.Args):
    group = args.String(fetch_options=fetch_groups)
    pass


@provider.grant()
def grant(p: BasicProvider, subject, args):
    pass


provider = BasicProvider(config_loader=provider.NoopLoader())

runtime = AWSLambdaRuntime(provider=provider, args_cls=Args)


def test_lambda_handler_works():
    event = {
        "type": "grant",
        "data": {"subject": "testuser", "args": {"group": "test"}},
    }
    runtime.handle(event=event, context=None)


def test_options_works():
    event = {
        "type": "options",
        "data": {"arg": "group"},
    }
    runtime.handle(event=event, context=None)


def test_provider_describe():
    event = {"type": "describe"}
    return runtime.handle(event=event, context=None)



from commonfate_provider import provider, diagnostics, args


class BasicProvider(provider.Provider):
    pass


class ExampleProvider(provider.Provider):
    value = provider.String()


@provider.config_validator(name="List Users")
def can_list_users(provider: ExampleProvider, diagnostics: diagnostics.Logs) -> None:
    diagnostics.info("some message here")


@provider.config_validator(name="Fails")
def fails(provider: ExampleProvider, diagnostics: diagnostics.Logs) -> None:
    raise Exception("something bad happened")


def test_init_works():
    config = '{"value": "test"}'
    got = ExampleProvider(provider.StringLoader(config))
    assert got.value.get() == "test"


def test_export_schema_works():
    got = ExampleProvider.export_schema()
    want = {"value": {"type": "string"}}
    assert got == want


def test_capabilities_works():
    got = provider.capabilities(_internal_key="BasicProvider")
    want = {"builtin": {}}
    assert got == want


def test_provider_config_validation_works():
    config = '{"value": "test"}'
    prov = ExampleProvider(provider.StringLoader(config))
    got = prov.validate_config()
    want = {
        "can_list_users": {
            "logs": [{"level": "info", "msg": "some message here"}],
            "success": True,
        },
        "fails": {
            "logs": [{"level": "error", "msg": "something bad happened"}],
            "success": False,
        },
    }
    assert got == want



import typing
from commonfate_provider import resources


class FirstResource(resources.Resource):
    """
    An example resource
    """

    value: str
    related_field: typing.Optional[str] = resources.Related("FirstResource")


class SecondResource(resources.Resource):
    value: str
    related_field: typing.Optional[str] = resources.Related(FirstResource)


@resources.fetcher
def load_resources():
    pass


def test_audit_schema_works():
    got = resources.audit_schema()

    want = {
        "resourceLoaders": {"load_resources": {"title": "load_resources"}},
        "resources": {
            "FirstResource": {
                "description": "An example resource",
                "properties": {
                    "id": {"title": "Id", "type": "string"},
                    "related_field": {
                        "relatedTo": "FirstResource",
                        "title": "Related Field",
                        "type": "string",
                    },
                    "value": {"title": "Value", "type": "string"},
                },
                "required": ["id", "value", "related_field"],
                "title": "FirstResource",
                "type": "object",
            },
            "SecondResource": {
                "properties": {
                    "id": {"title": "Id", "type": "string"},
                    "related_field": {
                        "relatedTo": "FirstResource",
                        "title": "Related Field",
                        "type": "string",
                    },
                    "value": {"title": "Value", "type": "string"},
                },
                "required": ["id", "value", "related_field"],
                "title": "SecondResource",
                "type": "object",
            },
        },
    }
    assert got == want


def test_build_resource_graph_works():
    pass
    # resources.set_fixture(
    #     [
    #         FirstResource(id="1", value="test"),
    #         SecondResource(id="2", related_field="1", value="test"),
    #     ]
    # )
    # got = resources.graph()
    # want = ["FirstResource:1", "SecondResource:2"]

    # assert got.vs()["name"] == want

    # edges = [e.vertex_tuple for e in got.es()]
    # edge_names = [(source["name"], target["name"]) for (source, target) in edges]

    # assert edge_names == [("FirstResource:1", "SecondResource:2")]



import typing
import toml
from commonfate_provider.provider import Provider
from commonfate_provider.args import Args
import importlib.util
import importlib
import os


def load_provider(cwd: str = "") -> typing.Tuple[typing.Type[Provider], typing.Type[Args]]:
    """
    Loads the Common Fate Provider by reading the provider.toml file.
    This method dynamically instantiates the provider class.
    """
    l = os.path.join(cwd,"provider.toml")
    config = toml.load(l)

    language = config["language"]

    if language != "python3.9":
        # this wrapping code is written in Python, so we can only handle
        # Providers written in Python.
        raise Exception(f"invalid language: {language}")

    provider_class = load_class(cwd,config["python"]["class"])
    arg_class = load_class(cwd,config["python"]["arg_schema"])
    return (provider_class, arg_class)


def load_class(cwd:str,path: str):
    """
    Dynamically loads a class. The path argument is the path
    to the class in the format 'module.ClassName'
    """
    components = path.split(".")  # [module, ClassName]
    modulePathWithinProviderFolder = "/".join(components[:-1])+".py"
    
    # Import the module using an absolute path to the python file containing the Provider class
    spec = importlib.util.spec_from_file_location('provider',os.path.join(cwd,modulePathWithinProviderFolder))
    # Loading the module

    if spec is None or spec.loader is None:
        raise Exception("expected spec and spec.loader not to be None")
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    class_name = components[-1]  # ClassName
    my_class = getattr(module, class_name)
    return my_class
